<?php

class apcCache extends Cache {
  /**
   * page_fast_cache
   *   This tells CacheRouter to use page_fast_cache.
   *
   *   @return bool TRUE
   */
  function page_fast_cache() {
    return TRUE;
  }

  /**
   * get()
   *   Return item from cache if it is available.
   *
   * @param string $key
   *   The key to fetch.
   * @return mixed object|bool
   *   Returns either the cache object or FALSE on failure
   */
  function get($key) {
    $cache = parent::get($key);
    if (isset($cache)) {
      return $cache;
    }

    $cache = apc_fetch($this->key($key));
    if (is_object($cache)) {
      if ($cache->serialized) {
        $cache->data = unserialize($cache->data);
      }
    }
    else {
      $cache = 0;
    }
    parent::set($key, $cache);
    return $cache;
  }

  /**
   * set()
   *   Add item into cache.
   *
   * @param string $key
   *   The key to set.
   * @param string $value
   *   The data to store in the cache.
   * @param string $expire
   *   The time to expire in seconds.
   * @param string $headers
   *   The page headers.
   * @return bool
   *   Returns TRUE on success or FALSE on failure
   */
  function set($key, $value, $expire = CACHE_PERMANENT, $headers = NULL) {
    if (empty($key)) {
      return FALSE;
    }
    $return = FALSE;
    // Create new cache object.
    $cache = new stdClass;
    $cache->cid = $key;
    $cache->created = time();
    $cache->expire = $expire;
    $cache->headers = $headers;

    if (!is_string($value)) {
      $cache->serialized = TRUE;
      $cache->data = serialize($value);
    }
    else {
      $cache->serialized = FALSE;
      $cache->data = $value;
    }

    //Store temporary as permanent - clean in flush
    if ($expire == CACHE_TEMPORARY || $expire == CACHE_PERMANENT) {
      $ttl = 0;
    }
    else {
      $ttl = $expire - $cache->created;
    }

    // Get full key for storage and set it to $expire so we can keep track of the bin
    $full_key = $this->key($key);

    // Attempt to store full key and value
    if (apc_store($full_key, $cache, $ttl)) {
      // Update static cache
      parent::set($key, $cache);
      $return = TRUE;

      // Get lookup table to be able to keep track of bins
      $this->lock();
      $lookup = apc_fetch($this->lookup);

      // If the lookup table is empty, initialize table
      if (empty($lookup)) {
        $lookup = array();
      }

      $lookup[$full_key] = $expire;

      // Resave the lookup table (even on failure)
      apc_store($this->lookup, $lookup);

      // Remove lock.
      $this->unlock();
    }
    return $return;
  }

  /**
   * delete()
   *   Remove item from cache.
   *
   * @param string $key
   *   The key to delete.
   * @return mixed object|bool
   *   Returns either the cache object or FALSE on failure
   */
  function delete($key) {
    if (substr($key, -1, 1) == '*') {
      //Clear static array cache (no wildcards)
      parent::flush();

      $this->lock();
      $lookup = apc_fetch($this->lookup);
      if (is_array($lookup)) {
        if ($key == '*') {
          //Fast clean of lookup
          apc_store($this->lookup, array());
          $this->unlock();
          foreach ($lookup as $k => $v) {
            apc_delete($k);
          }
          return TRUE;
        }
        else {
          $key = $this->key(substr($key, 0, strlen($key) - 1));
          foreach ($lookup as $k => $v) {
            if (strpos($k, $key) === 0) {
              apc_delete($k);
              unset($lookup[$k]);
            }
          }
        }
      }
      else {
      	$lookup = array();
      }
      apc_store($this->lookup, $lookup);
      $this->unlock();
    }
    else {
      //Remove only key - clean $lookup on flush
      parent::delete($key);
      return apc_delete($this->key($key));
    }
  }

  /**
   * flush()
   *   Flush the entire cache.
   *
   * @param none
   * @return mixed bool
   *   Returns TRUE
   */
  function flush() {
    parent::flush();
    $this->lock();

    // Get lookup table to be able to keep track of bins
    $lookup = apc_fetch($this->lookup);

    // If the lookup table is empty, remove lock and return
    if (empty($lookup)) {
      $this->unlock();
      return TRUE;
    }

    // Cycle through keys and remove each entry from the cache
    foreach ($lookup as $k => $v) {
      if ($v == CACHE_TEMPORARY) {
        apc_delete($k);
        unset($lookup[$k]);
      }
      elseif (apc_fetch($k) === FALSE) {
        unset($lookup[$k]);
      }
    }

    // Resave the lookup table (even on failure)
    apc_store($this->lookup, $lookup);

    // Remove lock
    $this->unlock();

    return TRUE;
  }

  /**
   * lock()
   *   lock the cache from other writes.
   *
   * @param none
   * @return string
   *   Returns TRUE on success, FALSE on failure
   */
  function lock() {
    if (function_exists('apc_add')) {
      // Lock once by trying to add lock file, if we can't get the lock, we will loop
      // for 3 seconds attempting to get lock.  If we still can't get it at that point,
      // then we give up and return FALSE.
      if (apc_add($this->lock, TRUE) === FALSE) {
        $time = time();
        while (apc_add($this->lock, TRUE) === FALSE) {
          if (time() - $time >= 3) {
            return FALSE;
          }
        }
      }
    }
    else {
      //For old versions of apc (before 3.0.13)
    	$time = time();
    	while (apc_fetch($this->lock)) {
        if (time() - $time >= 3) {
          return FALSE;
        }
    	}
    	apc_store($this->lock, TRUE);
    }
    return TRUE;
  }

  /**
   * unlock()
   *   lock the cache from other writes.
   *
   * @param none
   * @return bool
   *   Returns TRUE on success, FALSE on failure
   */
  function unlock() {
    return apc_delete($this->lock);
  }
}